﻿#virtualSynthDriver.py
#A part of NonVisual Desktop Access (NVDA)
#This file is covered by the GNU General Public License.
#See the file COPYING for more details.
#Copyright 2011 World Light Information Limited and Hong Kong Blind Union.

from collections import OrderedDict
import config
import globalVars
import queueHandler
from logHandler import log
from synthSettingsRing import SynthSettingsRing
from speechDictHandler import loadVoiceDict
from speech import IndexCommand
import synthDriverHandler

class _SavedSettings(object):
	"""Stores original settings of a physical synthesizer."""

	def __init__(self,synth):
		self.params={
			'rate':synth.rate,
			'pitch':synth.pitch,
			'inflection':synth.inflection,
			'volume':synth.volume
		}
		self.endIndex=0xfffe

	def adjustParams(self,synth,rate,pitch,inflection,volume):
		"""Make relative changes to the parameters of a synthsizer instance.

		All relative parameters are ranged between 0 and 100.  0 means half of
		the original value;  50 means original value;  100 means twice the
		original value.
		@param synth: Synthesizer that requires parameter update.
		@type synth: L{SynthDriver}
		@param rate: Relative rate.
		@type rate: int
		@param pitch: Relative pitch.
		@type pitch: int
		@param inflection: Relative inflection.
		@type inflection: int
		@param volume: Relative volume.
		@type volume: int
		"""
		synth.rate=self._relativeToActual(rate,self.params['rate'])
		synth.pitch=self._relativeToActual(pitch,self.params['pitch'])
		synth.inflection=self._relativeToActual(inflection,self.params['inflection'])
		synth.volume=self._relativeToActual(volume,self.params['volume'])

	@classmethod
	def _relativeToActual(cls,relative,oldValue):
		"""Converts a percentage relatively.
		@param relative: The relative adjustment.  Ranges between 0 and 100.
		@type relative: int
		@param oldValue: The original percentage value.
		@type oldValue: int
		"""
		if relative<0:
			return oldValue
		elif relative<50:
			min=oldValue/2
			max=oldValue
			percent=relative*2
			return int(round(float(percent)/100*(max-min)+min))
		elif relative==50:
			return oldValue
		elif relative<=100:
			min=oldValue
			max=oldValue*2
			if max>100: max=100
			percent=(relative-50)*2
			return int(round(float(percent)/100*(max-min)+min))
		else:
			return oldValue

class VirtualSynthDriver(synthDriverHandler.SynthDriver):
	"""Abstract virtual synthesizer that manages multiple drivers.

	Support for rate, pitch, inflection and volume is provided by this class.
	At a minimum, multi-language synthesizer drivers inheriting from this class must override the L{speak} methods.
	The L{speak} method should break the text into portions and feed each portion to a different synthesizer by calling L{speakPortion}.
	"""
	synths={}
	settings={}
	curSynth=None
	genID=None
	queue=[]
	_lastIndex=None

	# Default parameters
	relativeRate=50
	relativePitch=50
	relativeInflection=50
	relativeVolume=50

	def isSupported(self,settingName):
		if settingName in ("rate", "pitch", "inflection", "volume"):
			return True
		return super(VirtualSynthDriver,self).isSupported(settingName)

	def saveSettings(self):
		super(VirtualSynthDriver,self).saveSettings()
		conf=config.conf["speech"][self.name]
		conf["rate"]=self.relativeRate;
		conf["pitch"]=self.relativePitch;
		conf["inflection"]=self.relativeInflection;
		conf["volume"]=self.relativeVolume;

	def loadSettings(self):
		conf=config.conf["speech"][self.name]
		conf["rate"]=self.relativeRate;
		conf["pitch"]=self.relativePitch;
		conf["inflection"]=self.relativeInflection;
		conf["volume"]=self.relativeVolume;
		super(VirtualSynthDriver,self).loadSettings()

	def pause(self,switch):
		if self.curSynth in self.synths:
			self.synths[self.curSynth].pause(switch)

	def cancel(self):
		del self.queue[:]
		if self.curSynth in self.synths:
			self.synths[self.curSynth].cancel()
		self.curSynth=None

	def terminate(self):
		"""Terminates this virtual synthesizer.
		Subclasses should call the superclass method first.
		"""
		self.cancel()
		if self.genID!=None:
			queueHandler.cancelGeneratorObject(self.genID)
		self.unloadAllSynthsExcept()

	def _get_lastIndex(self):
		return self._lastIndex

	def _get_rate(self):
		return self.relativeRate

	def _set_rate(self,rate):
		self.relativeRate=rate

	def _get_pitch(self):
		return self.relativePitch

	def _set_pitch(self,pitch):
		self.relativePitch=pitch

	def _get_inflection(self):
		return self.relativeInflection

	def _set_inflection(self,inflection):
		self.relativeInflection=inflection

	def _get_volume(self):
		return self.relativeVolume

	def _set_volume(self,volume):
		self.relativeVolume=volume

	def _processPortion(self):
		"""Speaks next text portion in queue with the requested synthesizer.
		@return: C{True} if speech was started, C{False} if not.
		@rtype: bool
		"""
		(name,portion,index,rate,pitch,inflection,volume)=self.queue.pop(0)
		self._lastIndex=0 if index==None else index
		if name in self.synths:
			self.curSynth=name
			synth=self.synths[name]
			setting=self.settings[name]
			setting.adjustParams(synth,rate,pitch,inflection,volume)
			setting.endIndex=setting.endIndex^1
			# Feed speech sequence to target synthesizer
			speechSequence=[]
			if index is not None:
				speechSequence.append(IndexCommand(index))
			speechSequence.append(portion)
			speechSequence.append(IndexCommand(setting.endIndex))
			speechSequence.append(" ")
			synth.speak(speechSequence)
			return True
		else:
			self.curSynth=None
			return False

	def _portionGenerator(self):
		while True:
			if self.curSynth!=None:
				# Check whether current synthersizer has finished speaking
				if not self.curSynth in self.synths:
					self.curSynth=None
				elif self.synths[self.curSynth].lastIndex==self.settings[self.curSynth].endIndex:
					self.curSynth=None
			if self.curSynth==None and len(self.queue)>0:
				# Process next portion when there is no active synth
				self._processPortion()
			yield

	def loadSynth(self,name):
		"""Adds a new synthesizer to the list of managed synthersizers.
		@return: C{True} if the synthesizer is now managed, C{False} if not.
		@rtype: bool
		@param name: The name of the synthesizer that needs to be managed.
		@type name: str
		"""
		if len(name)==0:
			# Ignore blank names
			return False
		if name==self.name:
			# Prevent recursive loading of synthesizer
			return False
		if name in self.synths:
			# Open each synthesizer only once
			return True
		try:
			newSynth=synthDriverHandler._getSynthDriver(name)()
			if name in config.conf["speech"]:
				newSynth.loadSettings()
			else:
				# Create the new section.
				config.conf["speech"][name]={}
				if newSynth.isSupported("voice"):
					voice=newSynth.voice
				else:
					voice=None
				synthDriverHandler.changeVoice(newSynth,voice)
				newSynth.saveSettings()
			# Start managing synthesizer now
			self.synths[name]=newSynth
			self.settings[name]=_SavedSettings(newSynth)
			# Speak dummy string to setup last index
			speechSequence=[]
			speechSequence.append(IndexCommand(self.settings[name].endIndex))
			speechSequence.append(" ")
			newSynth.speak(speechSequence)
			return True
		except:
			log.error("Error while importing SynthDriver %s"%name,exc_info=True)
			return False

	def unloadAllSynthsExcept(self,retained=()):
		"""Unloads all synthesizers managed by this driver with exception.
		@param retained: Names of synthesizer to retain under management.
		@type retained: list or tuple of str
		"""
		self.cancel()
		for name in self.synths.keys():
			if name in retained: continue
			self.synths[name].terminate()
			del self.synths[name]
			del self.settings[name]

	def setMasterSynth(self,name):
		"""Uses a synthesizer as the master and initializes NVDA accordingly.
		@param name: The name of the master synthesizer.
		@type name: str
		"""
		if name in self.synths:
			synth=self.synths[name]
			if synth.isSupported("voice"):
				#We need to call changeVoice here so voice dictionaries can be managed
				synthDriverHandler.changeVoice(synth,synth.voice)
			else:
				#start or update the synthSettingsRing (for those synths which do not support 'voice')
				if globalVars.settingsRing:
					globalVars.settingsRing.updateSupportedSettings(synth)
				else:
					globalVars.settingsRing=SynthSettingsRing(synth)
				loadVoiceDict(synth)

	def getSynthSetting(self,name,attr):
		"""Obtains the setting of a managed synthesizer
		@return: Value of the required setting.
		@rtype: int
		@param name: The name of the synthesizer.
		@type name: str
		@param attr: The name of the setting.
		@type attr: str
		"""
		if name in self.synths:
			setting=self.settings[name]
			if attr in setting.params:  return setting.params[attr]
		return 50

	def setSynthSetting(self,name,attr,value):
		"""Changes the setting of a managed synthesizer
		@param name: The name of the synthesizer.
		@type name: str
		@param attr: The name of the setting.
		@type attr: str
		@param value: New value for the setting.
		@type value: int
		"""
		if name in self.synths:
			setting=self.settings[name]
			if attr in setting.params:  setting.params[attr]=value

	def speakPortion(self,name,portion,index):
		"""Queues a text portion to be spoken by a particular synthesizer.
		@param name: Name of the synthesizer that will speak the text portion.
		@type name: str
		@param portion: The text portion to be spoken.
		@type portion: str
		@param index: An index to associate with this text portion.
		@type index: int
		"""
		if self.genID==None:
			self.genID=queueHandler.registerGeneratorObject(self._portionGenerator())
		if len(portion)>0:
			self.queue.append([
				name,
				portion,
				index,
				self.relativeRate,
				self.relativePitch,
				self.relativeInflection,
				self.relativeVolume
			])
